#define _XOPEN_SOURCE 600
#include "io.h"
#include "addr.h"
#include "crypto.h"
#include "log.h"
#include "util.h"
#include "string.h"
#include "def.h"
#include "config.h"
#include <ftw.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>

//TODO: chek errno osv
bool file_exists(char const *path){
  return access(path,F_OK) == 0;
}

bool file_touch(char const *path){
  FILE *f = file_open(path,"ab");
  if(!f) return 0;
  if(!file_close(f)) return 0;
  disk_written(path);
  return 1;
}

bool file_delete(char const *path){
  if(remove(path) == 0){
#ifdef LOG_IOSPAM
    log_out("file deleted: '%s'",path);
#endif
    return 1;
  }else{
    log_errno("remove");
    irclog_err("failed to delete file: '%s'",path);
    return 0;
  }
}

bool file_rename(char const *from,char const *to){
  if(rename(from,to) == 0){
#ifdef LOG_IOSPAM
    log_out("file renamed: '%s' -> '%s'",from,to);
#endif
    return 1;
  }else{
    log_errno("rename");
    irclog_err("failed to rename file: '%s' -> '%s'",from,to);
    return 0;
  }
}

bool file_stat(char const *path,struct stat *sb){
  if(stat(path,sb) == 0){
#ifdef LOG_IOSPAM
    log_out("file stated: '%s'",path);
#endif
    return 1;
  }else{
    log_errno("stat");
    irclog_err("failed to stat file: '%s'",path);
    return 0;
  }
}

FILE *file_open(char const *path,char const *mode){
  FILE *f = fopen(path,mode);
  if(!f){
    log_errno("fopen");
    irclog_err("failed to open file: '%s'",path);
  }
  return f;
}

bool file_close(FILE *f){
  int r = fclose(f);
  if(r != 0){
    log_errno("fclose");
    irclog_err("failed to close file handle");
    return 0;
  }
  return 1;
}

bool file_seek(FILE *f,size_t off,int origin){
  if(fseek(f,off,origin) != 0){
    log_errno("fseek");
    irclog_err("failed to seek in file");
    return 0;
  }
  return 1;
}

bool file_read(FILE *f,void *data,size_t size){
  size_t r = fread(data,1,size,f);
  if(r != size){
    if(ferror(f)){
      log_errno("fread");
      irclog_err("failed to read from file");
      return 0;
    }else{
      irclog_err("unexpected eof reached");
      return 0;
    }
  }
  return 1;
}

bool file_write(FILE *f,void const *data,size_t size){
  size_t r = fwrite(data,1,size,f);
  if(r != size){
    log_errno("fwrite");
    irclog_err("failed to write to file");
    return 0;
  }
  return 1;
}

bool file_gets(FILE *f,void *buf,size_t buf_size){
  char *r = fgets(buf,buf_size,f);
  if(!r && ferror(f)){
    log_errno("fgets");
    irclog_err("failed to read from file");
  }
  return r;
}

bool file_printf(FILE *f,char const *fmt,...){
  va_list args;
  int     r;

  va_start(args,fmt);
  r = vfprintf(f,fmt,args);
  va_end(args);
  if(r < 0){
    log_errno("fprintf");
    irclog_err("failed to write to file");
    return 0;
  }
  return 1;
}

bool mkdir_p(char const *path){
  char  buf[PATH_BUF];
  char *p = buf;
  char  t;

  strzcpy(buf,path,sizeof(buf));
  do{
    p   = strchr0(p,'/');
    t   = *p;
    *p  = 0;
    if(mkdir(buf,DATA_MODE) == 0){
#ifdef LOG_IOSPAM
      log_out("directory created: '%s'",buf);
#endif
    }else if(errno != EEXIST){
      log_errno("mkdir");
      irclog_err("failed to create directory: '%s'",path);
      return 0;
    }
    *p  = t;
    p   = strnotchr0(++p,'/');
  }while(t);
  return 1;
}

static int ftw_rm_rf(char const         *path,
                     struct stat const  *sb,
                     int                 flag,
                     struct FTW         *ftw){

  (void)sb;
  (void)flag;
  (void)ftw;
  return file_delete(path) ? 0 : -1;
}

bool rm_rf(char const *path){
  if(nftw(path,ftw_rm_rf,64,FTW_DEPTH | FTW_PHYS) == 0){
    return 1;
  }else{
    log_errno("nftw");
    irclog_err("failed to recursively delete directory: '%s'",path);
    return 0;
  }
}

bool str_store(char const *path,char const *str){
  FILE   *f;
  size_t  str_len;

  f = file_open(path,"wb");
  if(!f) return 0;

  str_len = strlen(str);
  if(!file_write(f,str,str_len)){
    file_close(f);
    return 0;
  }

  if(!file_write(f,"\n",1)){
    file_close(f);
    return 0;
  }

  if(!file_close(f)) return 0;
  disk_written(path);
  return 1;
}

bool str_load(char const *path,char *buf,size_t buf_size){
  FILE         *f;
  size_t        size;
  struct stat   sb;

  if(!file_stat(path,&sb)) return 0;
  size = (size_t)sb.st_size;

  if(size+1 > buf_size){
    irclog_err("file size + terminator exceeds "
               "buffer size: '%s' (%zu > %zu)",
               path,size+1,buf_size);
    return 0;
  }

  f = file_open(path,"rb");
  if(!f) return 0;

  if(!file_read(f,buf,size)){
    file_close(f);
    return 0;
  }

  ((uint8_t *)buf)[size] = 0;
  str_stripnl(buf);

  file_close(f);
  disk_read(path);
  return 1;
}

bool int_store(char const *path,int64_t i){
  char buf[INT64_BUF];
  snprintf(buf,sizeof(buf),"%ld",i);
  return str_store(path,buf);
}

bool int_load(char const *path,int64_t *i){
  char buf[INT64_LEN+WHITE_LEN+1+1]; //int\n\0
  if(!str_load(path,buf,sizeof(buf))) return 0;
  *i = strtoll(buf,0,10);
  return 1;
}

bool bool_store(char const *path,bool b){
  return str_store(path,b ? "1" : "0");
}

bool bool_load(char const *path,bool *b){
  char    buf[INT64_LEN+WHITE_LEN+1+1]; //int\n\0
  int64_t i;
  if(!str_load(path,buf,sizeof(buf))) return 0;
  i   = strtoll(buf,0,10);
  *b  = i != 0;
  return 1;
}

bool hash_store(char const *path,hash_t const *h,size_t cnt){
  FILE *f;

  f = file_open(path,"wb");
  if(!f) return 0;

  for(size_t x = 0;x < cnt;x++){
    char hash_b64[BASE64_ENC_SIZE(sizeof(hash_t),1)];
    base64_enc(h[x],sizeof(hash_t),hash_b64,sizeof(hash_b64));

    if(!file_printf(f,"%s\n",hash_b64)){
      file_close(f);
      return 0;
    }
  }

  if(!file_close(f)) return 0;
  disk_written(path);
  return 1;
}

static bool hash_from_b64(hash_t hash,char const *b64_hash){
  size_t n;

  if(!str_valid(b64_hash,HASH_B64_LEN,HASH_B64_LEN,BASE64_CHARSET)){
    return 0;
  }

  n = base64_dec(b64_hash,hash,HASH_SIZE,0);
  assert(n == HASH_SIZE); (void)n;
  return 1;
}

bool hash_load(char const *path,hash_t *h,size_t *cnt,size_t max){
  FILE   *f;
  char    buf[HASH_B64_LEN+WHITE_LEN+1+1]; //base64(hash)\n\0
  char   *b;
  size_t  n;

  f = file_open(path,"rb");
  if(!f) return 0;

  n = 0;
  while(file_gets(f,buf,sizeof(buf))){
    b = str_trim(buf);
    if(!*b) continue;

    if(n >= max){
      log_err("file '%s' max entries reached (%zu)",path,max);
      file_close(f);
      return 0;
    }

    if(!hash_from_b64(h[n],b)){
      log_err("file '%s' contains invalid hash: '%s'",path,b);
      file_close(f);
      return 0;
    }

    n++;
  }

  if(ferror(f)){
    file_close(f);
    return 0;
  }

  file_close(f);
  disk_read(path);
  *cnt = n;
  return 1;
}

bool pass_store(char const *path,uint8_t const *pass){
  char pass_b64[B64_PASS_BUF];
  base64_enc(pass,PASS_SIZE,pass_b64,sizeof(pass_b64));
  return str_store(path,pass_b64);
}

bool pass_load(char const *path,uint8_t *pass){
  char    buf[B64_PASS_LEN+WHITE_LEN+1+1]; //base64(pass_hash)\n\0
  char   *b;
  size_t  n;

  if(!str_load(path,buf,sizeof(buf))){
    return 0;
  }

  b = str_trim(buf);

  if(!str_valid(b,B64_PASS_LEN,B64_PASS_LEN,BASE64_CHARSET)){
    log_err("file '%s' contains invalid password hash: '%s'",path,b);
    return 0;
  }

  n = base64_dec(b,pass,PASS_SIZE,0);
  assert(n == PASS_SIZE); (void)n;

  return 1;
}

bool addr_store(char const *path,addr_t const *addr){
  char at[AT_BUF];
  at_from_addr(at,sizeof(at),addr);
  return str_store(path,at);
}

bool addr_load(char const *path,addr_t *addr){
  char  buf[AT_LEN+WHITE_LEN+1+1]; //at\n\0
  char *b;

  if(!str_load(path,buf,sizeof(buf))){
    return 0;
  }

  b = str_trim(buf);
  if(*b && !addr_from_at(addr,b)){
    log_err("file '%s' contains invalid address: "
            "'%s' (format is ip:port)",path,b);
    return 0;
  }

  return 1;
}

void disk_read(char const *path){
  (void)path;
#ifdef LOG_IOSPAM
  log_out("file read: '%s'",path);
#endif
}

void disk_written(char const *path){
  (void)path;
#ifdef LOG_IOSPAM
  log_out("file written: '%s'",path);
#endif
}

void disk_unsync(void){
  irclog_out("ATTENTION - DISK OUT OF SYNC!");
}

